/* --- Copyright University of Sussex 1994. All rights reserved. ----------
 > File:			$poplocal/local/lib/winved.p
 > Purpose:			Terminal-style Ved running in a single Windows window
 > Author:			Robert John Duncan, Jul 25 1994
 */

compile_mode:pop11 +strict;

section $-winved =>
	vedwindowskeys
	vedwindowsscreen
	winved_main
	winved
;

include vedfile_struct;

    /*	Windows messages, keycodes, etc. These must agree with #define
		values in
		    <windows.h>
		    "popwin.h"
    */

define :inline MAKE_POP_MESSAGE(n);
	(WM_USER + 16:100 + (n))
enddefine;

define :inline MAKE_POP_EDIT_MESSAGE(n);
	MAKE_POP_MESSAGE(100 + (n))
enddefine;

constant macro (
	
	/* Windows Messages */
	
	WM_USER	= 16:400,
	
	/* Pop Messages */

	POPM_CLOSE					= MAKE_POP_MESSAGE(1),
	POPM_CONFIRM_EXIT			= MAKE_POP_MESSAGE(2),
	POPM_EDIT_COMMAND			= MAKE_POP_EDIT_MESSAGE(`\e`),
	POPM_EDIT_CHAR				= MAKE_POP_EDIT_MESSAGE(257),
	POPM_EDIT_FUNCTION			= MAKE_POP_EDIT_MESSAGE(258),
	POPM_EDIT_SIZE 				= MAKE_POP_EDIT_MESSAGE(259),
	POPM_EDIT_MOUSE				= MAKE_POP_EDIT_MESSAGE(260),

	/* Windows Virtual Key codes */
	
	VK_UP		= 16:26,
	VK_DOWN		= 16:28,
	VK_LEFT		= 16:25,
	VK_RIGHT	= 16:27,
	
	VK_INSERT	= 16:2D,
	VK_DELETE	= 16:2E,
	VK_HOME		= 16:24,
	VK_END		= 16:23,
	VK_PRIOR	= 16:21,
	VK_NEXT		= 16:22,
		
	VK_F1		= 16:70,
	VK_F2		= 16:71,
	VK_F3		= 16:72,
	VK_F4		= 16:73,
	VK_F5		= 16:74,
	VK_F6		= 16:75,
	VK_F7		= 16:76,
	VK_F8		= 16:77,
	VK_F9		= 16:78,
	VK_F10		= 16:79,
	VK_F11		= 16:7A,
	VK_F12		= 16:7B,
	VK_F13		= 16:7C,
	VK_F14		= 16:7D,
	VK_F15		= 16:7E,
	VK_F16		= 16:7F,
	VK_F17		= 16:80,
	VK_F18		= 16:81,
	VK_F19		= 16:82,
	VK_F20		= 16:83,
	VK_F21		= 16:84,
	VK_F22		= 16:85,
	VK_F23		= 16:86,
	VK_F24		= 16:87,
	
	VK_NUMLOCK	= 16:90,
);

constant macro (
	WINVED_OUTPUT_BUFFER_SIZE = 128,
);

vars
	winved_base_window = false,
	winved_edit_window = false,
	
	winved_procs_waiting = [],
	winved_chars_waiting = [],
	
	winved_output_buffer = false,
	winved_output_index = 0,
	winved_insert_mode = false,
;

define winved_qq;
	ved_save_file_globals();
	lvars file, needs_confirm = false;
	for file in vedbufferlist do
		if file(VF_CHANGED) and (file(VF_WRITEABLE) or vedwriteallfiles) then
			true -> needs_confirm;
			quitloop;
		endif;
	endfor;
	if needs_confirm then
		lvars message = consPopWindowMessage(POPM_CONFIRM_EXIT, 0, 0);
		lvars dosave = popwin_send_message(winved_base_window, message);
		if dosave == 0 then
			;;; YES
			vedwritefiles();
		elseif dosave == 2 then
			;;; CANCEL
			return;
		endif;
	endif;
	ved_rqq();
enddefine;

	/*	Function keymap: converts Windows virtual key codes to character
		sequences observed by Ved. Each sequence is preceded by ESC [ X
		where X is:
			0	for key alone
			1	for ctrl + key
			2	for shift + key
			3	for shift + ctrl + key
		So, for example, 'Up Arrow' maps to: ESC [ 0 1 1
		while Ctrl + 'Up Arrow' maps to: ESC [ 1 1 1
	*/
define winved_keymap =
	newproperty([
	
		;;; Arrow keys
		[^VK_UP			'11']
		[^VK_DOWN		'12']
		[^VK_LEFT		'13']
		[^VK_RIGHT		'14']
	
		;;; Named keys
		[^VK_INSERT		'21']
		[^VK_DELETE		'22']
		[^VK_HOME		'23']
		[^VK_END		'24']
		[^VK_PRIOR		'25']
		[^VK_NEXT		'26']
		
		;;; Function Keys
		[^VK_F1			'51']
		[^VK_F2			'52']
		[^VK_F3			'53']
		[^VK_F4			'54']
		[^VK_F5			'55']
		[^VK_F6			'56']
		[^VK_F7			'57']
		[^VK_F8			'58']
		[^VK_F9			'59']
		[^VK_F10		'60']
		[^VK_F11		'61']
		[^VK_F12		'62']
		[^VK_F13		'63']
		[^VK_F14		'64']
		[^VK_F15		'65']
		[^VK_F16		'66']
		[^VK_F17		'67']
		[^VK_F18		'68']
		[^VK_F19		'69']
		[^VK_F20		'70']
		[^VK_F21		'71']
		[^VK_F22		'72']
		[^VK_F23		'73']
		[^VK_F24		'74']
	
		;;; Keypad prefix; will be followed by ASCII key / * - + <CR>
		[^VK_NUMLOCK	'?']
	], 64, false, "perm");
enddefine;

	/*	Get next input from the Ved window - char or procedure
	*/
define lconstant read_input(chars_only);
	;;; check for buffered input
	unless chars_only or winved_procs_waiting == [] then
		return(sys_grbg_destpair(winved_procs_waiting) -> winved_procs_waiting);
	elseunless winved_chars_waiting == [] then
		return(sys_grbg_destpair(winved_chars_waiting) -> winved_chars_waiting);
	endunless;
	;;; allow for timeout
	if isinteger(pop_timeout_secs) then
		define raise_timeout();
			pop_timeout();
		enddefine;
		dlocal % sys_timer(raise_timeout) % = pop_timeout_secs * 1e6;
	endif;
	;;; get next message from the editor window
	lconstant message = writeable consPopWindowMessage(0,0,0);
	repeat
	    popwin_get_message(winved_edit_window, message) -> ;
	    ;;; dispatch on the message type
	    lvars type = popWindowMessageType(message);
	    if type == POPM_EDIT_CHAR then
		    ;;; character key press: character code is in the message
		    return(popWindowMessageArg(message));
	    elseif type == POPM_EDIT_FUNCTION then
		    ;;; function key press: map key code to a character sequence
			lvars code = popWindowMessageArg(message), chars;
	        if winved_keymap(code fi_&& 16:FF) ->> chars then
				[% `[`, (code fi_>> 8) fi_+ `0`, explode(chars) %]
					-> winved_chars_waiting;
				return(`\e`);
	        endif;
		elseif type == POPM_EDIT_SIZE then
			;;; window resized:
			;;; 	nrows = LOWORD(arg)
			;;; 	ncols = HIWORD(arg)
            define lconstant resize(nrows, ncols);
	            lvars nrows, ncols;
	            unless nrows == vedscreenlength
				and ncols == vedscreenwidth
				then
			        ;;; stop the bell ringing on every size change
			        dlocal vedscreenbell = identfn;	
		            vedresize(nrows, ncols);
	            endunless;
            enddefine;
			lvars code = popWindowMessageArg(message);
			lvars proc = resize(% code && 16:FFFF, (code >> 16) && 16:FFFF %);
			returnunless(chars_only)(proc);
			winved_procs_waiting nc_<> [^proc] -> winved_procs_waiting;
		elseif type == POPM_EDIT_MOUSE then
			;;; mouse press:
			;;;		row = LOWORD(arg)
			;;;		col = HIWORD(arg)
			define lconstant follow_mouse(row, col);
				if ved_on_status then vedstatusswitch() endif;
				if row < vedscreenoffset or row >= vedscreenoffset+vedwindowlength then
					;;; other window: may mishap if there's no other file
					vedswapfiles();
				endif;
				row - vedscreenoffset -> row;
				if row == 0 then
					;;; status line
					unless ved_on_status then vedstatusswitch() endunless;
					max(1, col + 1 - vedwcolumnoffset) -> vedcolumn;
				else
					vedjumpto(
						vedlineoffset + row,
						vedcolumnoffset + max(1, col + 1 - vedwcolumnoffset));
				endif;
			enddefine;
			lvars code = popWindowMessageArg(message);
			lvars col = (code >> 16) && 16:FFFF, row = code && 16:FFFF;
			lvars proc = follow_mouse(%row, col%);
			returnunless(chars_only)(proc);
			winved_procs_waiting nc_<> [^proc] -> winved_procs_waiting;
		elseif type == POPM_CLOSE then
			;;; base window closing -- quit Ved
			returnunless(chars_only)(winved_qq);
			winved_procs_waiting nc_<> [^winved_qq] -> winved_procs_waiting;
	    ;;; else ignore
	    endif;
	endrepeat;
enddefine;

define winved_read_input();
	read_input(false);
enddefine;

	/*	Get next ASCII character from the window
	*/
define winved_read_ascii();
	read_input(true);
enddefine;

	/*	Test for input waiting
	*/
define winved_input_waiting();
	;;; this doesn't take account of messages in the queue
	winved_chars_waiting /== [] or winved_procs_waiting /== [];
enddefine;

	/*	Clear input waiting
	*/
define winved_clear_input();
	;;; likewise
	[] ->> winved_chars_waiting -> winved_procs_waiting;
enddefine;

	/*	Send a message to the edit window
	*/
define winved_send_message(message);
	lvars message;
	popwin_send_message(winved_edit_window, message) -> ;
enddefine;

	/*	Flush the output buffer
	*/
define winved_flush_output();
	lconstant message = consPopWindowMessage(POPM_EDIT_COMMAND, 0, 0);
	unless winved_output_index == 0 then
	    winved_output_index -> popWindowMessageArg(message);
	    winved_output_buffer -> popWindowMessageExtra(message);
	    winved_send_message(message);
	    0 -> winved_output_index;
	endunless;
enddefine;

	/*	Copy substring to the output buffer
	*/
constant procedure ( winved_char_out );

define winved_substring_out(string, csub, nchars);
	if isstring16(string) then
		fast_repeat nchars times
			winved_char_out(fast_subscrs(csub, string));
			csub fi_+ 1 -> csub;
		endrepeat;
	else
	    lvars n = WINVED_OUTPUT_BUFFER_SIZE fi_- winved_output_index;
	    if n == 0 then
		    winved_flush_output();
		    WINVED_OUTPUT_BUFFER_SIZE -> n;
	    endif;
	    repeat
			if nchars fi_< n then nchars -> n endif;
		    move_bytes(
				csub, string,
				winved_output_index fi_+ 1, winved_output_buffer,
				n);
		    winved_output_index fi_+ n -> winved_output_index;
		    nchars fi_- n -> nchars;
	    quitif(nchars == 0);
		    winved_flush_output();
		    csub fi_+ n -> csub;
		    WINVED_OUTPUT_BUFFER_SIZE -> n;
	    endrepeat;
	endif;
enddefine;

	/*	Copy char or string to the output buffer
	*/
define winved_char_out(char);
	lvars char;
	if isinteger(char) then
		if winved_output_index == WINVED_OUTPUT_BUFFER_SIZE then
			winved_flush_output();
		endif;
		winved_output_index fi_+ 1 -> winved_output_index;
		if char fi_> 255 then `?` -> char endif;
	    char -> fast_subscrs(winved_output_index, winved_output_buffer);
	else
		;;; assume string
		winved_substring_out(char, 1, datalength(char));
	endif;
enddefine;

	/*	Send a control sequence with row/column coordinates
	*/
define winved_control(code, x, y);
	lvars code, x, y;
	if x fi_> 255 or y fi_> 255 then
		;;; need to use long output form
		lconstant message = consPopWindowMessage(0, 0, 0);
		winved_flush_output();
		MAKE_POP_EDIT_MESSAGE(code) -> popWindowMessageType(message);
		((y fi_<< 16) fi_|| x) -> popWindowMessageArg(message);
		winved_send_message(message);
	else
	    if winved_output_index fi_> #_< WINVED_OUTPUT_BUFFER_SIZE - 3 >_#
		then
		    winved_flush_output();
	    endif;
	    winved_output_index fi_+ 3 -> winved_output_index;
	    code -> fast_subscrs(winved_output_index fi_- 2, winved_output_buffer);
	    x -> fast_subscrs(winved_output_index fi_- 1, winved_output_buffer);
	    y -> fast_subscrs(winved_output_index, winved_output_buffer);
	endif;
enddefine;

	/*	Position the caret in the edit window
	*/
define winved_goto(col, line);
	lvars col, line;
	winved_control(`\^Y`, col fi_- 1, line fi_- 1);
	col -> vedscreencolumn;
	line -> vedscreenline;
enddefine;

	/*	Set the scrolling region
	*/
define winved_setscrollregion(top, bottom);
	lvars top, bottom;
	winved_control(`\^Z`, top fi_- 1, bottom);
	1000 -> vedscreencolumn;
	1000 -> vedscreenline;
enddefine;

	/*	Initialise globals
	*/
define winved_init();
	unless winved_output_buffer then
		init_fixed(WINVED_OUTPUT_BUFFER_SIZE, string_key) -> winved_output_buffer;
	endunless;
	unless winved_base_window then
	    popwin_create_base_window("WinVed") -> winved_base_window;
	    popWindowChildren(winved_base_window)(1) -> winved_edit_window;
	endunless;
enddefine;

	/*	Set screen control variables for Windows
	*/
define vedwindowsscreen();

	;;; create windows
	winved_init();
	
	;;; set terminal name
	"windows" -> vedterminalname;
	
	;;; redefine low-level I/O procedures
	winved_read_input -> vedscr_read_input;
	winved_read_ascii -> vedscr_read_ascii;
	winved_input_waiting -> vedscr_input_waiting;
	winved_clear_input -> vedscr_clear_input;
	winved_char_out -> vedscr_char_out;
	winved_substring_out -> vedscr_substring_out;
	winved_flush_output -> vedscr_flush_output;

	;;; set initial screen size, flags, etc.
	80 -> vedscreenwidth;
	24 -> vedscreenlength;
	false -> vedscreenwrap;
	false -> vednomoveinsert;
	
	;;; set screen control sequences
	`\^P` -> vvedscreencharup;
	`\^N` -> vvedscreenchardown;
	`\^B` -> vvedscreencharleft;
	`\^F` -> vvedscreencharright;
	`\^M` -> vvedscreenscreenleft;
	`\^L` -> vvedscreenclear;
	`\^K` -> vvedscreencleartail;
	`\^D` -> vvedscreendeletechar;
	`\^U` -> vvedscreendeleteline;
	`\^O` -> vvedscreeninsertline;
	`\^G` -> vvedscreenbell;
	`\^S` -> vvedscreenscrollup;
	`\^T` -> vvedscreenscrolldown;
	`\^W` -> vvedscreeninsertmode;
	`\^X` -> vvedscreenovermode;
	
	winved_goto -> vedscreenxy;
	winved_setscrollregion -> vedsetscrollregion;

	identfn ->> vedscreenreset -> vedrestorescreen;
	
	;;; special characters
	;;; `\(151)` -> vedscreenstatus_-_mark;
	
enddefine;

	/*	Set key bindings for Windows
		(NB: shifted keys left unbound -- should do mark+move)
	*/
define vedwindowskeys();
	vedset keys
		
		chardelete		= ^H

		screenbell		= ^X				;;; Reserved for CUT
		screenbell		= ^C				;;; Reserved for COPY
		screenbell		= ^V				;;; Reserved for PASTE
		
		charup			= esc [ 0 1 1		;;; Up Arrow
		chardown		= esc [ 0 1 2		;;; Down Arrow
		charleft		= esc [ 0 1 3		;;; Left Arrow
		charright		= esc [ 0 1 4		;;; Right Arrow
		
		scrolldown<>charup= esc [ 1 1 1		;;; Ctrl Up Arrow
		scrollup<>chardown= esc [ 1 1 2		;;; Ctrl Down Arrow
		wordleft		= esc [ 1 1 3		;;; Ctrl Left Arrow
		wordright		= esc [ 1 1 4		;;; Ctrl Right Arrow
		
		charuplots		= esc esc [ 0 1 1	;;; Esc Up Arrow
		chardownlots	= esc esc [ 0 1 2	;;; Esc Down Arrow
		charleftlots	= esc esc [ 0 1 3	;;; Esc Left Arrow
		charrightlots	= esc esc [ 0 1 4	;;; Esc Right Arrow
		
		ENTER static	= esc [ 0 2 1		;;; Insert
		dotdelete		= esc [ 0 2 2		;;; Delete
		screenleft		= esc [ 0 2 3		;;; Home
		textright		= esc [ 0 2 4		;;; End
		prevscreen		= esc [ 0 2 5		;;; Page Up
		nextscreen		= esc [ 0 2 6		;;; Page Down
		
		topfile			= esc [ 1 2 3		;;; Ctrl Home
		endfile			= esc [ 1 2 4		;;; Ctrl End
		screenup		= esc [ 1 2 5		;;; Ctrl Page Up
		screendown		= esc [ 1 2 6		;;; Ctrl Page Down
		
		textleft		= esc esc [ 0 2 3	;;; Esc Home
		screenright		= esc esc [ 0 2 4	;;; Esc End

		clearhead		= esc [ 0 5 2			;;; F2
		linedelete		= esc [ 0 5 3			;;; F3
		cleartail		= esc [ 0 5 4			;;; F4
		"ENTER yankw"	= esc esc [ 0 5 2		;;; Esc F2
		"ENTER yankl"	= esc esc [ 0 5 3		;;; Esc F3
		"ENTER yankw"	= esc esc [ 0 5 4		;;; Esc F4

		wordleftdelete	= esc [ 0 5 5			;;; F5
		wordrightdelete	= esc [ 0 5 6			;;; F6
		marklo			= esc [ 0 5 7			;;; F7
		markhi			= esc [ 0 5 8			;;; F8
		"ENTER yankw"	= esc esc [ 0 5 5		;;; Esc F5
		"ENTER yankw"	= esc esc [ 0 5 6		;;; Esc F6
		"ENTER mbf"		= esc esc [ 0 5 7		;;; Esc F7
		"ENTER mef"		= esc esc [ 0 5 8		;;; Esc F8
	
		"ENTER m"		= esc [ 0 5 9			;;; F9
		"ENTER t"		= esc [ 0 6 0			;;; F10
		"ENTER mi"		= esc esc [ 0 5 9		;;; Esc F9
		"ENTER ti"		= esc esc [ 0 6 0		;;; Esc F10

		"enterkey"		= esc [ 0 ? CR			;;; KP Enter
		"statusswitch"	= esc [ 0 ? +			;;; KP +
		"redocommand"	= esc [ 0 ? -			;;; KP -
		"loadline"		= esc [ 0 ? *			;;; KP *
		"ENTER hkeys"	= esc [ 0 ? /			;;; KP /

	endvedset;
enddefine;

vars winved = true;

;;; -----------------------------------------------------------------------

include subsystem

	/*	Start up procedure for GUI Poplog
	*/
define winved_start();
compile_mode :vm -prmprt;
	returnunless(sys_fname_nam(poparg0) = 'winpop11');

	lvars save_poparglist = (poparglist, [] -> poparglist);
	lvars first_call = true;
	
    define main_loop with_props false;
        dlocal 0 %,
		    if dlocal_context == 1
		    or vedbufferlist == [] and ved_char_in_stream == []
		    then
			    sysexit();
		    else
			    ;;; restart
			    true -> pop_first_setpop;
		    endif
	    %;
		
        define nowrite_defaults();
	        false -> vedwriteable;
	        vedveddefaults();
        enddefine;

        define start(im);
	        vedscreenlength / 2 -> vedstartwindow;
			if im then
		        ved_clear();
		        sys_subsystems_init(SS_BANNER);
		        chain(vedsetpop);
			endif;
        enddefine;

		if first_call then
            lvars subss = subsystem or "pop11";
			lvars type = subscr_subsystem(SS_FILE_EXTN, subss);
            current_directory dir_>< ('output' <> type)
	            ->> ved_chario_file
	            ->> vedlmr_print_in_file
	            ->  vedlmr_errs_in_file;
	        veduseterm(%"windows"%) <> erase -> vedinitterm;
            vedscreenlength -> vedstartwindow;
            false -> subsystem;
			false -> first_call;
			if save_poparglist == [] then
				vedinput(start(%true%));
		        vededit(ved_chario_file, nowrite_defaults);
			else
				vedinput(start(%false%));
				lvars files = save_poparglist;
				[] -> save_poparglist;
				until (dest(files) ->> files) == [] do
					lvars file = ();
					vedinput(vededit(%file%));
				enduntil;
		        vededit();
			endif;
	    elseif vedbufferlist == [] then
	        vededit(ved_chario_file, nowrite_defaults);
	    endif;
	    until vedbufferlist == [] and ved_char_in_stream == [] do
	        ved_ved();
	    enduntil;
    enddefine;

    sysCOMPILE(% main_loop %) -> pop_setpop_compiler;
    true -> pop_first_setpop;
    setpop -> interrupt;
enddefine;
;;;
sys_runtime_apply(winved_start);

endsection;		/* $-winved */
